/*
 * file CqRecord.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.cq.CqRecord
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.wvcm.stp.cq;
import static com.ibm.rational.wvcm.stpex.StpExBase.PROPERTY_NAMESPACE;

import java.util.Date;
import java.util.List;

import javax.wvcm.Feedback;
import javax.wvcm.PropertyNameList;
import javax.wvcm.ResourceList;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;

import com.ibm.rational.wvcm.stp.StpActivity;
import com.ibm.rational.wvcm.stp.StpErrorMessageException;
import com.ibm.rational.wvcm.stp.StpProperty;
import com.ibm.rational.wvcm.stp.StpResource;
import com.ibm.rational.wvcm.stpex.StpExBase;
import com.ibm.rational.wvcm.stpex.StpExEnumeration;
import com.ibm.rational.wvcm.stpex.StpExEnumerationBase;

/**
 * A CqRecord resource proxy provides access to a record in a database, possibly
 * through its change context.
 * 
 * <p>
 * To ensure the integrity of record field data, the modification of a record is
 * a multi-step process.
 * 
 * <ul>
 * <li>First, a writable copy of the record is added to the database's change
 * context using the current field values from the database. This writable copy
 * of the record is visible only in the provider where it was created. While it
 * exists in the change context, it hides the record in the database as seen
 * through the provider.
 * <li>Field values are modified in the CqRecord proxy by the client.
 * <li>As needed, modified field values are written to the writable copy of the
 * record on the server. (see
 * {@link CqContextResource#doWriteProperties(Feedback, List) doWriteProperties}.)
 * <li>The client modifies a record or a set of records in this fashion until
 * all have been modified and made consistent to the client's satisfaction. Then
 * they are written back to the database is a single atomic operation. (See
 * {@link com.ibm.rational.wvcm.stp.cq.CqUserDb#doDeliver(Feedback, List)} or
 * {@link com.ibm.rational.wvcm.stp.cq.CqContextResource#doDeliver(Feedback)} or
 * {@link CqContextResource#doWriteProperties(Feedback, List) doWriteProperties}.)
 * </ul>
 * 
 * If these steps succeed, the client can be assured that the data written back
 * to the record is faithfully derived from the data read from the record.
 * 
 * <p>
 * If doWriteProperties is applied to a record that is not yet writable, an
 * action must be started under which the record will be modified. If an action
 * is not specified as the value of the ACTION property of this proxy, the
 * default MODIFY-type action for the record is used.
 * 
 * <p>
 * The fields of a record manifest themselves as properties of the record
 * resource and thus record field values are made available to the client via
 * the normal property request mechanism. The names, types, and other
 * characteristics of the fields of a record are available as properties of the
 * {@link CqFieldDefinition CqFieldDefinition} resources enumerated by the
 * CqRecordType resource associated with the record.
 * 
 * <p>
 * In addition to their value, record fields also have meta-properties such as
 * their requiredness, their original value, and a list of their legal values.
 * Each of these can be requested using a MetaPropertyName object in a
 * nested PropertyRequest of the field PropertyName.
 * <p>
 * The user-friendly specification for the location of a record has the form
 * 
 * <pre>
 *  <b>cq.record:</b><i>&lt;record-type&gt;</i>/<i>&lt;record-id&gt;</i>@<i>&lt;db-set&gt;</i>/<i>&lt;user-db&gt;</i>
 * </pre>
 * 
 * @see CqRecordType
 * @see CqFieldDefinition
 * @see CqFieldValue
 */
public interface CqRecord extends StpActivity, CqContextResource
{
    /**
     * Defines the value of the specified property for this proxy. The
     * PropertyName may specify a record field name, in which case the Object
     * may be a {@link CqFieldValue CqFieldValue} structure or the actual value
     * of the field.
     * 
     * @see StpResource#setProperty(PropertyNameList.PropertyName, Object)
     */
    <T> void setProperty(PropertyName<T> name,
                         T value);

    /**
     * Returns the value defined by this proxy for the property specified by the
     * given PropertyName. The PropertyName may specify a record field, in which
     * case the Object returned is <i>not</i> a
     * {@link CqFieldValue CqFieldValue} structure, but the actual value of the
     * field. (Clients may use {@link #getField getField} or
     * {@link StpResource#getMetaProperties(PropertyName)} to obtain a
     * CqFieldValue structure if the additional information provided by the
     * CqFieldValue structure is needed.).
     * 
     * @see javax.wvcm.Resource#getProperty(javax.wvcm.PropertyNameList.PropertyName)
     */
    <T> T getProperty(PropertyName<T> name) throws WvcmException;

    /**
     * A CqRecordType proxy for the record-type resource that defines what kind
     * of record this is.
     */
    PropertyName<CqRecordType> RECORD_TYPE =
        new PropertyName<CqRecordType>(StpExBase.PROPERTY_NAMESPACE,
                                       "record-type");

    /**
     * A CqRecordType proxy for the record-type resource that defines what kind
     * of record this is.
     * 
     * @return A proxy for the record-type of this record is returned.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #RECORD_TYPE RECORD_TYPE} property.
     */
    CqRecordType getRecordType() throws WvcmException;

    /**
     * A subclass of PropertyName used for naming the record resource properties
     * that are schema-defined fields of a ClearQuest record. Note that, as with
     * PropertyName, the type parameter specifies the field's value type.
     */
    public static class FieldName<T> extends PropertyName<T>
    {
        /**
         * Constructs a PropertyName designating the resource property that
         * corresponds to a field of a ClearQuest record.
         * 
         * @param fieldName The schema-defined name of the record field. Must
         *            not be null.
         */
        public FieldName(String fieldName)
        {
            super(StpExBase.FIELD_NAMESPACE, fieldName);
        }
    }

    /**
     * Returns the meta-property bundle of a record field defined by this proxy.
     * Since CqFieldValue is an extension of StpProperty, this method is merely
     * a type-specialized version of
     * {@link StpResource#getMetaProperties(javax.wvcm.PropertyNameList.PropertyName)}.
     * 
     * @param fieldName The name of the desired field expressed as a FieldName
     *            object. Cannot be <code>null</code>.
     * 
     * @return A CqFieldValue structure is returned containing the field
     *         value(s) and/or attributes defined by this proxy.
     * 
     * @throws WvcmException If the field value denoted by the given field name
     *             is not defined by this proxy--that is, if no field values or
     *             attributes were requested or could not be retrieved when
     *             requested.
     */
    <T> CqFieldValue<T> getFieldInfo(FieldName<T> fieldName)
        throws WvcmException;

    /**
     * Defines a new meta-property bundle for a record field in this proxy. Only
     * the VALUE meta-property of the CqFieldValue is written to the resource,
     * but the TYPE meta-property is often useful to the library for
     * interpreting the VALUE when expressed as a String image.
     * 
     * @param fieldName The name of the field to be updated expressed as a
     *            FieldName object. Cannot be <code>null</code>.
     * 
     * @param value A CqFieldValue object containing the new field value. Must
     *            not be <b>null</b>
     */
    <T> void setFieldInfo(FieldName<T> fieldName,
                          CqFieldValue<T> value);

    /**
     * Defines in this proxy, a new value for a field of this record. The field
     * will not be updated until this proxy is used to invoke a "do" method.
     * 
     * @param fieldName The name of the field whose value is to be set
     * @param value The new value for the field expressed as a native Java
     *            object. The String image of the value may be used.
     */
    void setField(String fieldName,
                  Object value);

    /**
     * Retrieves the value of a field previously requested by an instance of
     * FieldName having the same fieldName parameter as this method. To retrieve
     * other meta-properties requested for the field, use
     * {@link #getFieldInfo(com.ibm.rational.wvcm.stp.cq.CqRecord.FieldName)}.
     * 
     * @param fieldName The name of the field for which a value was requested.
     * @return The value of the requested field returned by the server expressed
     *         as a native Java object (not just its String image).
     * @throws WvcmException If the field value was not requested or could not
     *             be retrieved.
     */
    Object getField(String fieldName) throws WvcmException;

    /**
     * Creates a new record at the location specified by this proxy, initialized
     * with the dirty properties of this proxy. The type of record created is
     * determined by the parent location of this proxy (which must name an
     * existing record type resource).
     * <p>
     * An action must be started under which the record will be created. If an
     * action is not specified as the value of the ACTION property of this
     * proxy, the default SUBMIT-type action for the record is used.
     * <p>
     * The server is permitted to modify the location as it deems fit. The proxy
     * returned by this method is a proxy that addresses the location actually
     * chosen by the server.
     * 
     * @param feedback A Feedback object requesting the properties desired from
     *            the record created by this operation. May be <b>null</b> if
     *            no properties are desired.
     * 
     * @return An CqRecord proxy for the created record. This proxy will define
     *         the properties specified in the feedback argument.
     * 
     * @throws WvcmException If the record cannot be created with the given
     *             properties.
     */
    CqRecord doCreateRecord(Feedback feedback) throws WvcmException;

    /**
     * Creates a new record at the location specified by this proxy, initialized
     * with the dirty properties of this proxy. The type of record created is
     * determined by the parent location of this proxy (which must name an
     * existing record type resource).
     * <p>
     * An action must be started under which the record will be created. If an
     * action is not specified as the value of the ACTION property of this
     * proxy, the default SUBMIT-type action for the record is used.
     * <p>
     * The server is permitted to modify the location as it deems fit. The proxy
     * returned by this method is a proxy that addresses the location actually
     * chosen by the server.
     * 
     * @param feedback A Feedback object requesting the properties desired from
     *            the record created by this operation. May be <b>null</b> if
     *            no properties are desired.
     * 
     * @param deliveryOrder If {@link CqProvider#HOLD} the created record is
     *            left in a writable state in the change context--the new record
     *            in the change context must be delivered before the it becomes
     *            visible to other providers. If not {@link CqProvider#HOLD},
     *            the modified and moribund resources specified by this
     *            parameter will be delivered to or deleted from the database in
     *            the order indicated. To deliver all modified and moribund
     *            resources in an arbitrary order, use
     *            {@link CqProvider#DELIVER_ALL}. To deliver just this new
     *            record, use {@link CqProvider#DELIVER}. Must not be <b>null</b>.
     * @return A CqRecord proxy for the created record. This proxy will define
     *         the properties specified in the feedback argument.
     * 
     * @throws WvcmException If the record cannot be created with the given
     *             properties.
     */
    CqRecord doCreateRecord(Feedback feedback,
                            List<CqContextResource> deliveryOrder)
        throws WvcmException;

    /**
     * Executes a hook (record script) named in the NAMED_HOOK_LIST property of
     * this record's RECORD_TYPE resource. The impact on the change context in
     * which the hook is executed is essentially the same as executing
     * {@link com.ibm.rational.wvcm.stp.cq.CqContextResource#doWriteProperties(Feedback, List)}
     * in that context (where the hook is in control of which properties are
     * written).
     * <p>
     * As with doWriteProperties, the record need not be editable prior to
     * invoking this method, but, unlike doWriteProperties, if it isn't editable
     * and the hook wants to modify it, then the hook must be prepared to start
     * an action itself to do so. An implicit Modify-type action is not
     * automatically started for the hook (except in the case noted below).
     * <p>
     * If this proxy contains updated property values when this method is
     * invoked, an attempt will be made to write those new property values to
     * the server <i>before</i> the hook is executed. If such an update is
     * required and this record isn't editable, it will be made editable using
     * the default Modify-type action and the hook will be applied to that
     * editable version of the record before delivering the changes to the
     * database. If the property update fails, this operation fails immediately
     * and subsequent execution of the hook will not be attempted.
     * 
     * @param hook A CqHook proxy for the named hook to be fired. If the client
     *            knows the name of the hook to be fired, it should construct a
     *            CqHook proxy at the computed location <i>record</i><code>.stpLocation().recomposeWithNamespace(Namespace.HOOK).parent().child(</code><i>hook-name</i><code>)</code>
     *            If the client does not know the name of the hook to be fired,
     *            a list of available hooks to choose from can be obtained from
     *            the NAMED_HOOK_LIST property of the parent record type of this
     *            record.
     * @param parameters A String containing any parameters required by the
     *            hook. The format and content of this parameter string is
     *            prescribed solely by the hook, which is also responsible for
     *            parsing it into a usable value or values.
     * @param feedback Properties to be retrieved from the target record and/or
     *            those records changed by the execution of the hook (and after
     *            delivery if delivery has been requested).
     * @param deliveryOrder Delivery of the changes made to the record will be
     *            considered only if the hook succeeds. If this argument is
     *            {@link CqProvider#HOLD} the modified record is left in a
     *            writable state in the change context--the modified record must
     *            be delivered before the changes made by the hook become
     *            visible outside the change context. If not
     *            {@link CqProvider#HOLD}, the modified and moribund resources
     *            specified by this parameter will be delivered to or deleted
     *            from the database in the order indicated. To deliver all
     *            modified and moribund resources in an arbitrary order, use
     *            {@link CqProvider#DELIVER_ALL}. To deliver just this modified
     *            record, use {@link CqProvider#DELIVER}. Must not be <b>null</b>.
     * @return A CqRecord proxy for the modified record, populated with the
     *         resources requested by the feedback argument.
     * @throws WvcmException if there are problems finding or executing the
     *             named hook, writing updated properties, or starting the
     *             implied action.
     * @throws StpErrorMessageException if the hook returns a non-empty result
     */
    CqRecord doFireNamedHook(CqHook hook,
                             String parameters,
                             Feedback feedback,
                             List<CqContextResource> deliveryOrder)
        throws WvcmException, StpErrorMessageException;

    /**
     * Causes the execution of an action of type
     * CqAction.Type.RECORD_SCRIPT_ALIAS. Such actions are defined solely by a
     * hook script, which is executed when the action is applied to a record.
     * When this method is called, the action is started on this record and the
     * action's hook script is executed as the sole content of that action and
     * then the action is committed. The impact of this operation on the change
     * context is essentially equivalent to executing in the same context
     * {@link com.ibm.rational.wvcm.stp.cq.CqContextResource#doWriteProperties(Feedback, List)}
     * where the deliveryOrder argument forces an immediate delivery of the
     * change context.
     * <p>
     * Because of this prescribed processing
     * <ul>
     * <li>This record must not be editable when this method is called so that
     * the specified RECORD_SCRIPT_ALIAS action can be started on it.
     * <li> This record must not contain updated property values because only
     * the record script of the action is allowed to make modifications to the
     * record once the action has started.
     * <li>If the action script succeeds, delivery of the targeted record is
     * immediate and cannot be deferred--i.e., CqProvider.HOLD <i>may not</i>
     * be used for the delivery order list. Furthermore, the targeted record is
     * delivered first, regardless of the content of the delivery order list.
     * <li>If the action script fails, no delivery will be attempted and the
     * change context will remain as it was before invoking this method.
     * </ul>
     * <p>
     * Typically, schema designers have associated arbitrary executable code
     * with a record type, represented in a GUI as a button with the display
     * name of the record type action.
     * 
     * @param action A CqAction for an action of type RECORD_SCRIPT_ALIAS. This
     *            action must be defined by the record type of this record
     *            instance.
     * @param feedback A Feedback object requesting the properties desired from
     *            resources involved in this operation. May be <b>null</b> if
     *            no properties are desired.
     * @param deliveryOrder Upon successful execution of the action script, the
     *            target record will be delivered to its database. Then, any
     *            other resources specified by this argument will be delivered.
     *            Must not be <b>null</b> and must not be CqProvider.HOLD.
     * @return A String containing the return value computed by the hook script.
     * @throws WvcmException if there are problems
     *             <ul>
     *             <li>finding the record,
     *             <li>finding a RECORD_SCRIPT_ALIAS as specified
     *             <li>starting the action
     *             <li>executing the script
     *             <li>delivering the changes to the database
     *             </ul>
     */
    String doFireRecordScriptAlias(CqAction action,
                                   Feedback feedback,
                                   List<CqContextResource> deliveryOrder)
        throws WvcmException;

    /**
     * Returns the date of the submit action in the record history field.
     * 
     * @see javax.wvcm.Resource#getCreationDate()
     */
    Date getCreationDate() throws WvcmException;

    /**
     * Returns the date of the last action applied to the record from its
     * history
     * 
     * @see javax.wvcm.Resource#getLastModified()
     */
    Date getLastModified() throws WvcmException;

    /**
     * Returns the login name of the user who submitted the record
     * 
     * @see com.ibm.rational.wvcm.stp.StpResource#getCreatorLoginName()
     */
    String getCreatorLoginName() throws WvcmException;

    /**
     * The error message generated when the record, in its current state, was
     * validated against the repository schema for compliance with its rules.
     * This property will be empty if the record is fully compliant (which is
     * always the case if the record is not open for edit).
     * <p>
     * When this property is requested for a record open for editing, the record
     * is validated before any other properties are retrieved. Thus, all
     * properties returned with this property reflect the state of the record
     * after validation was attempted.
     * <p>
     * Before a record being edited can be delivered, it must be validated (even
     * if no fields have been changed). Normally, validation is performed as the
     * first phase of a delivery and any validation errors cause the delivery to
     * be aborted. Requesting this property will force validation of the record
     * without attempting to deliver it. The time and resources needed to 
     * perform a record validation vary with the schema. Unless you have 
     * specific knowledge of the target database's schema, it is best to assume
     * that the validation process is an expensive operation and design your
     * application to request this property only when validation is really 
     * required.
     * <p>
     * If you are changing the contents of a record programmatically, you should
     * make sure that your code provides valid data.
     * <p>
     * You should not attempt to parse and interpret the returned String
     * programmatically, because the error text may change in future releases.
     * If you want to try to correct the value in an field with incorrect
     * values, you can use the INVALID_FIELDS property to discover which fields
     * are invalid. The MESSAGE_TEXT meta-property of those fields will identify
     * problems specific to the field.
     */
    PropertyName<String> VALIDATION_ERROR =
        new PropertyName<String>(StpExBase.PROPERTY_NAMESPACE,
                                 "validation-error");

    /**
     * Returns the value of the {@link #VALIDATION_ERROR VALIDATION_ERROR}
     * property as defined by this proxy.
     * 
     * @return Returns a String containing the error message generated when this
     *         record was validated against the repository schema. Will never be
     *         <b>null</b>, but may be empty is the record field values are
     *         fully compliant with the rules of the schema.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #VALIDATION_ERROR VALIDATION_ERROR } property.
     */
    String getValidationError() throws WvcmException;

    /**
     * The action under which this record is being modified; null if the record
     * is not being modified.
     */
    PropertyName<CqAction> ACTION =
        new PropertyName<CqAction>(StpExBase.PROPERTY_NAMESPACE, "action");

    /**
     * Returns the value of the {@link #ACTION ACTION} property as defined by
     * this proxy.
     * 
     * @return Returns an Action proxy representing the action associated with
     *         this record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ACTION ACTION} property.
     */
    CqAction getAction() throws WvcmException;

    /**
     * Defines a new value for the {@link #ACTION ACTION} property of this
     * proxy. Set this value prior to invoking doWriteProperties,
     * doCreateRecord, or doUnbindAll to specify an action other than the
     * default action specified for those operations. If this property is not
     * set prior to invocation of these methods, an action of the required type
     * will be used if one is unique defined.
     * 
     * @param action A CqAction proxy identifying the action under which this
     *            record is to be created, modified, or deleted. Must not be
     *            <b>null</b>. For the update to succeed, the given action must
     *            be on the LEGAL_ACTIONs list for this record.
     */
    CqRecord setAction(CqAction action);

    /**
     * The type of the action under which this record is being modified,
     * Type.NIL if the record is not being modified
     */
    PropertyName<CqAction.Type> ACTION_TYPE =
        new PropertyName<CqAction.Type>(StpExBase.PROPERTY_NAMESPACE,
                                        "action-type");

    /**
     * Returns the value of the {@link #ACTION ACTION} property as defined by
     * this proxy.
     * 
     * @return An Action.Type enumerator indicating the type of action
     *         associated with this record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ACTION_TYPE ACTION_TYPE} property.
     */
    CqAction.Type getActionType() throws WvcmException;

    /**
     * A list of {@link CqFieldValue CqFieldValue} objects corresponding to all
     * of the fields of this record. By default, no meta-property of any field
     * is included in the CqFieldValue's on this list. So, essentially, the
     * result is just a list of field property names. Using a NestedPropertyName
     * in the PropertyRequest of this property would allow the client to request
     * any or all meta-properties of the fields in the list.
     * <code>
     * new PropertyRequest(ALL_FIELD_VALUES.nest(VALUE.nest(VALUE))), 
     * </code>
     * for example, would request the value of each field of the record.
     */
    PropertyName<StpProperty.List<CqFieldValue<?>>> ALL_FIELD_VALUES =
        new PropertyName<StpProperty.List<CqFieldValue<?>>>(PROPERTY_NAMESPACE,
                                                            "all-field-values");

    /**
     * Returns the value of the {@link #ALL_FIELD_VALUES ALL_FIELD_VALUES}
     * property as defined by this proxy.
     * 
     * @return An StpProperty.List containing a CqFieldValue instance for each
     *         field of this record. Each CqFieldValue instance defines all of
     *         the metadata requested from the server for the field.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ALL_FIELD_VALUES ALL_FIELD_VALUES} property.
     */
    StpProperty.List<CqFieldValue<?>> getAllFieldValues() throws WvcmException;

    /** The default action associated with the current state of this record. */
    PropertyName<CqAction> DEFAULT_ACTION =
        new PropertyName<CqAction>(StpExBase.PROPERTY_NAMESPACE,
                                   "default-action");

    /**
     * Returns the value of the {@link #DEFAULT_ACTION DEFAULT_ACTION} property
     * as defined by this proxy.
     * 
     * @return A CqAction proxy representing the default action associated with
     *         this record in its current state. Will be null if a default
     *         action is not defined for this record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #DEFAULT_ACTION DEFAULT_ACTION} property.
     */
    CqAction getDefaultAction() throws WvcmException;

    /** Whether this record has been marked as a duplicate of another record. */
    PropertyName<Boolean> IS_DUPLICATE =
        new PropertyName<Boolean>(StpExBase.PROPERTY_NAMESPACE, "is-duplicate");

    /**
     * Returns the value of the {@link #IS_DUPLICATE IS_DUPLICATE} property as
     * defined by this proxy.
     * 
     * @return <b>true</b> if this record has been recorded as a duplicate of
     *         another record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #IS_DUPLICATE IS_DUPLICATE} property.
     * 
     * @see #getAllDuplicates()
     * @see #getDuplicates()
     * @see #getHasDuplicates()
     */
    boolean getIsDuplicate() throws WvcmException;

    /**
     * The CqRecord that is marked as the original of this duplicate record.
     */
    PropertyName<CqRecord> ORIGINAL =
        new PropertyName<CqRecord>(StpExBase.PROPERTY_NAMESPACE, "original");

    /**
     * Returns the value of the {@link #ORIGINAL ORIGINAL} property as defined
     * by this proxy.
     * 
     * @return A CqRecord proxy representing the record of which this is a
     *         duplicate. Will be <b>null</b> if this record is not a
     *         duplicate.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ORIGINAL ORIGINAL} property.
     */
    CqRecord getOriginal() throws WvcmException;

    /** The visible ID of this object's original record. */
    PropertyName<String> ORIGINAL_ID =
        new PropertyName<String>(StpExBase.PROPERTY_NAMESPACE, "original-id");

    /**
     * Returns the value of the {@link #ORIGINAL_ID ORIGINAL_ID} property as
     * defined by this proxy.
     * 
     * @return A String containing the record id for the record that this record
     *         is a duplicate of. Will be <b>null</b> if this record is not a
     *         duplicate.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ORIGINAL_ID ORIGINAL_ID} property.
     */
    String getOriginalId() throws WvcmException;

    /**
     * Whether other records have been marked as being duplicates of this record
     */
    PropertyName<Boolean> HAS_DUPLICATES =
        new PropertyName<Boolean>(StpExBase.PROPERTY_NAMESPACE,
                                  "has-duplicates");

    /**
     * Returns the value of the {@link #HAS_DUPLICATES HAS_DUPLICATES} property
     * as defined by this proxy.
     * 
     * @return <b>true</b> if records have been marked as duplicates of this
     *         record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #HAS_DUPLICATES HAS_DUPLICATES} property.
     * 
     * @see #getDuplicates()
     * @see #getAllDuplicates()
     * @see #getIsOriginal()
     */
    boolean getHasDuplicates() throws WvcmException;

    /**
     * True if this record has duplicates but is not itself a duplicate (i.e.
     * !IS_DUPLICATE & HAS_DUPLICATES)
     */
    PropertyName<Boolean> IS_ORIGINAL =
        new PropertyName<Boolean>(StpExBase.PROPERTY_NAMESPACE, "is-original");

    /**
     * Returns the value of the {@link #IS_ORIGINAL IS_ORIGINAL} property as
     * defined by this proxy.
     * 
     * @return <b>true</b> if this record has duplicates but is not, itself, a
     *         duplicate of another record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #IS_ORIGINAL IS_ORIGINAL} property.
     */
    boolean getIsOriginal() throws WvcmException;

    /** The immediate duplicates of this record. */
    PropertyName<ResourceList<CqRecord>> DUPLICATES =
        new PropertyName<ResourceList<CqRecord>>(StpExBase.PROPERTY_NAMESPACE,
                                                 "duplicates");

    /**
     * Returns the value of the {@link #DUPLICATES DUPLICATES} property as
     * defined by this proxy.
     * 
     * @return A ResourceList of CqRecord proxies representing all of the
     *         records that have been explicitly marked as a duplicate of this
     *         record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #DUPLICATES DUPLICATES} property.
     * 
     * @see #getAllDuplicates()
     */
    ResourceList<CqRecord> getDuplicates() throws WvcmException;

    /**
     * All of the duplicates of this record, including duplicates of duplicates,
     * etc.
     */
    PropertyName<ResourceList<CqRecord>> ALL_DUPLICATES =
        new PropertyName<ResourceList<CqRecord>>(StpExBase.PROPERTY_NAMESPACE,
                                                 "all-duplicates");

    /**
     * Returns the value of the {@link #ALL_DUPLICATES ALL_DUPLICATES} property
     * as defined by this proxy.
     * 
     * @return A ResourceList of CqRecord proxies representing all of the
     *         records marked as duplicates of this record or (recursively) one
     *         of its duplicates.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ALL_DUPLICATES ALL_DUPLICATES} property.
     */
    ResourceList<CqRecord> getAllDuplicates() throws WvcmException;

    /**
     * A list of the fields that were modified by the on-going action applied to
     * this record.
     */
    PropertyName<StpProperty.List<CqFieldValue<?>>> FIELDS_UPDATED_THIS_ACTION =
        new PropertyName<StpProperty.List<CqFieldValue<?>>>(StpExBase.PROPERTY_NAMESPACE,
                                                            "fields-updated-this-action");

    /**
     * Returns the value of the {@link #FIELDS_UPDATED_THIS_ACTION
     * FIELDS_UPDATED_THIS_ACTION} property as defined by this proxy.
     * 
     * @return An StpProperty.List containing a CqFieldValue for each field
     *         updated since the beginning of the current action on this record.
     *         Each CqFieldValue defines the metadata requested when this
     *         property was retrieved from the server.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #FIELDS_UPDATED_THIS_ACTION
     *             FIELDS_UPDATED_THIS_ACTION} property.
     */
    StpProperty.List<CqFieldValue<?>> getFieldsUpdatedThisAction()
        throws WvcmException;

    /**
     * A list of the fields that were modified under the current action since
     * the last time this property was set.
     * 
     * @see #setFieldsUpdatedThisGroup()
     */
    PropertyName<StpProperty.List<CqFieldValue<?>>> FIELDS_UPDATED_THIS_GROUP =
        new PropertyName<StpProperty.List<CqFieldValue<?>>>(StpExBase.PROPERTY_NAMESPACE,
                                                            "fields-updated-this-group");

    /**
     * Returns the value of the {@link #FIELDS_UPDATED_THIS_GROUP
     * FIELDS_UPDATED_THIS_GROUP} property as defined by this proxy.
     * 
     * @return An StpProperty.List containing a CqFieldValue instance for each
     *         field updated since the last time the
     *         {@link #FIELDS_UPDATED_THIS_GROUP} property was set.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #FIELDS_UPDATED_THIS_GROUP FIELDS_UPDATED_THIS_GROUP}
     *             property.
     */
    StpProperty.List<CqFieldValue<?>> getFieldsUpdatedThisGroup()
        throws WvcmException;

    /**
     * Clears the FIELDS_UPDATED_THIS_GROUP property. Only fields updated after
     * this value is set will be returned by subsequent requests for the
     * FIELDS_UPDATED_THIS_GROUP property.
     */
    void setFieldsUpdatedThisGroup();

    /**
     * The fields that were modified by the last field update.
     */
    PropertyName<StpProperty.List<CqFieldValue<?>>> FIELDS_UPDATED_THIS_SET_VALUE =
        new PropertyName<StpProperty.List<CqFieldValue<?>>>(StpExBase.PROPERTY_NAMESPACE,
                                                            "fields-updated-this-set-value");

    /**
     * Returns the value of the {@link #FIELDS_UPDATED_THIS_SET_VALUE
     * FIELDS_UPDATED_THIS_SET_VALUE} property as defined by this proxy.
     * 
     * @return An StpProperty.List containing a CqFieldValue instance for each
     *         field modified by the most recent client-initiated field update.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #FIELDS_UPDATED_THIS_SET_VALUE
     *             FIELDS_UPDATED_THIS_SET_VALUE} property.
     */
    StpProperty.List<CqFieldValue<?>> getFieldsUpdatedThisSetValue()
        throws WvcmException;

    /**
     * A list of the record's invalid fields.
     */
    PropertyName<StpProperty.List<CqFieldValue<?>>> INVALID_FIELDS =
        new PropertyName<StpProperty.List<CqFieldValue<?>>>(StpExBase.PROPERTY_NAMESPACE,
                                                            "invalid-fields");

    /**
     * Returns the value of the {@link #INVALID_FIELDS INVALID_FIELDS} property
     * as defined by this proxy.
     * 
     * @return An StpProperty.list of CqFieldValue instances identifying the
     *         fields of this record that are currently invalid.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #INVALID_FIELDS INVALID_FIELDS} property.
     */
    StpProperty.List<CqFieldValue<?>> getInvalidFields() throws WvcmException;

    /**
     * The actions that can be legally applied to this record in its current
     * state.
     */
    PropertyName<ResourceList<CqAction>> LEGAL_ACTIONS =
        new PropertyName<ResourceList<CqAction>>(StpExBase.PROPERTY_NAMESPACE,
                                                 "legal-actions");

    /**
     * Returns the value of the {@link #LEGAL_ACTIONS LEGAL_ACTIONS} property as
     * defined by this proxy.
     * 
     * @return A list of CqAction proxies representing the actions that are
     *         currently legal for this record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #LEGAL_ACTIONS LEGAL_ACTIONS} property.
     */
    ResourceList<CqAction> getLegalActions() throws WvcmException;

    /** The type (state-based or stateless) of the record. */
    PropertyName<CqRecord.Class> RECORD_CLASS =
        new PropertyName<CqRecord.Class>(StpExBase.PROPERTY_NAMESPACE,
                                         "record-class");

    /**
     * Returns the value of the {@link #RECORD_CLASS RECORD_CLASS} property as
     * defined by this proxy.
     * 
     * @return A CqRecord.Class enumeration identifying the class of this record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #RECORD_CLASS RECORD_CLASS} property.
     */
    Class getRecordClass() throws WvcmException;

    /**
     * The CqAttachmentFolder resources that correspond to the fields of this
     * record that have or may have attachments. The display name of the 
     * attachment folder is the same as the name of the attachment field.
     */
    PropertyName<ResourceList<CqAttachmentFolder>> ATTACHMENT_FOLDERS =
        new PropertyName<ResourceList<CqAttachmentFolder>>(StpExBase.PROPERTY_NAMESPACE,
                                                           "attachment-folders");

    /**
     * Returns the value of the {@link #ATTACHMENT_FOLDERS ATTACHMENT_FOLDERS}
     * property as defined by this proxy.
     * 
     * @return A ResourceList contain CqAttachmentFolder proxies identifying the
     *         attachment folders of this record.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #ATTACHMENT_FOLDERS ATTACHMENT_FOLDERS} property.
     */
    ResourceList<CqAttachmentFolder> getAttachmentFolders()
        throws WvcmException;

    /**
     * An enumeration of the possible types of records; state-based or
     * stateless.
     */
    public static enum Class implements StpExEnumeration
    {
        /** A record type having no state table. */
        STATELESS,

        /**
         * A record type whose records progress from state to state along
         * prescribed paths based on the actions applied to them.
         */
        STATE_BASED;
    }

    /** The name of the record's current state. */
    PropertyName<String> STATE_NAME =
        new PropertyName<String>(StpExBase.PROPERTY_NAMESPACE, "state-name");

    /**
     * Returns the value of the {@link #STATE_NAME STATE} property as defined by
     * this proxy.
     * 
     * @return A String containing the name of the state the record is currently
     *         in.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #STATE_NAME STATE} property.
     */
    String getStateName() throws WvcmException;

    /**
     * The replica that has mastership of this record.
     */
    PropertyName<CqReplica> CQ_MASTER_REPLICA =
        new PropertyName<CqReplica>(StpExBase.PROPERTY_NAMESPACE,
                                    "cq-master-replica");

    /**
     * Returns the value of the {@link #CQ_MASTER_REPLICA CQ_MASTER_REPLICA}
     * property as defined by this proxy.
     * 
     * @return A CqReplica proxy for the replica that has mastership of this
     *         record
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #CQ_MASTER_REPLICA CQ_MASTER_REPLICA} property.
     */
    CqReplica getCqMasterReplica() throws WvcmException;

    /**
     * Defines a new value for the {@link #CQ_MASTER_REPLICA CQ_MASTER_REPLICA}
     * property of this proxy.
     * 
     * @param master A CqReplica proxy identifying the replica that should
     *            assume mastership of this resource.
     */
    void setCqMasterReplica(CqReplica master);

    /**
     * A list of all the fields that were modified by the on-going action
     * applied to this record. Unlike the {@link #FIELDS_UPDATED_THIS_ACTION}
     * property, this list includes fields modified by the hooks that ran during
     * start up of the action.
     */
    PropertyName<StpProperty.List<CqFieldValue<?>>> FIELDS_UPDATED_THIS_ENTIRE_ACTION =
        new PropertyName<StpProperty.List<CqFieldValue<?>>>(StpExBase.PROPERTY_NAMESPACE,
                                                            "fields-updated-this-entire-action");

    /**
     * Returns the value of the
     * {@link #FIELDS_UPDATED_THIS_ENTIRE_ACTION FIELDS_UPDATED_THIS_ENTIRE_ACTION}
     * property as defined by this proxy.
     * 
     * @return An StpProperty.List containing a CqFieldValue for each field
     *         modified during the initialization and execution of the current
     *         action. Will never be <b>null</b>, but may be empty.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #FIELDS_UPDATED_THIS_ENTIRE_ACTION FIELDS_UPDATED_THIS_ENTIRE_ACTION}
     *             property.
     */
    StpProperty.List<CqFieldValue<?>> getFieldsUpdatedThisEntireAction()
        throws WvcmException;

    /**
     * A list of the fields of this record of
     * {@link CqFieldValue.ValueType#JOURNAL}
     */
    PropertyName<StpProperty.List<CqHistoryFieldValue>> HISTORY_FIELDS =
        new PropertyName<StpProperty.List<CqHistoryFieldValue>>(StpExBase.PROPERTY_NAMESPACE,
                                                                "history-fields");

    /**
     * Returns the value of the {@link #HISTORY_FIELDS HISTORY_FIELDS} property
     * as defined by this proxy.
     * 
     * @return An StpProperty.List containing a CqFieldValue instance for each
     *         field of the record whose field value type is JOURNAL. Will never
     *         be <b>null</b>, but may be empty.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #HISTORY_FIELDS HISTORY_FIELDS} property.
     */
    StpProperty.List<CqHistoryFieldValue> getHistoryFields()
        throws WvcmException;

    /**
     * A list of the actions that the current user may apply to this record.
     */
    PropertyName<ResourceList<CqAction>> LEGAL_ACCESSIBLE_ACTIONS =
        new PropertyName<ResourceList<CqAction>>(StpExBase.PROPERTY_NAMESPACE,
                                                 "legal-accessible-actions");

    /**
     * Returns the value of the
     * {@link #LEGAL_ACCESSIBLE_ACTIONS LEGAL_ACCESSIBLE_ACTIONS} property as
     * defined by this proxy.
     * 
     * @return A ResourceList of CqAction proxies enumerating the actions that
     *         the current user may apply to this record in its current state.
     *         Will never be <b>null</b>, but may be empty.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #LEGAL_ACCESSIBLE_ACTIONS LEGAL_ACCESSIBLE_ACTIONS}
     *             property.
     */
    ResourceList<CqAction> getLegalAccessibleActions() throws WvcmException;

    /**
     * The site-extension field for this record.
     */
    PropertyName<String> SITE_EXTENSION =
        new PropertyName<String>(StpExBase.PROPERTY_NAMESPACE, "site-extension");

    /**
     * Returns the value of the {@link #SITE_EXTENSION SITE_EXTENSION} property
     * as defined by this proxy.
     * 
     * @return A String containing the site-extension field used for this record
     *         when forming its site-extended name. Will never be <b>null</b>.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #SITE_EXTENSION SITE_EXTENSION} property.
     */
    String getSiteExtension() throws WvcmException;

    /**
     * A list of all the records of this record type whose names differ from the
     * name of this record only in their site extensions.
     */
    PropertyName<ResourceList<CqRecord>> SITE_EXTENDED_NAMES =
        new PropertyName<ResourceList<CqRecord>>(StpExBase.PROPERTY_NAMESPACE,
                                                 "site-extended-names");

    /**
     * Returns the value of the {@link #SITE_EXTENDED_NAMES SITE_EXTENDED_NAMES}
     * property as defined by this proxy.
     * 
     * @return A ResourceList containing a proxy for each CqRecord resource
     *         whose name differs from the name of this record only in the
     *         site extensions used.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #SITE_EXTENDED_NAMES SITE_EXTENDED_NAMES} property.
     */
    public ResourceList<CqRecord> getSiteExtendedNames() throws WvcmException;

    /**
     * The login name of the user currently holding a lock on this record ("*"
     * if the current user is not permitted to see user names). A zero-length
     * string indicates that the record is not locked. Setting this property to
     * the user's login-name or to a zero-length string locks or unlocks the
     * record, respectively.
     */
    PropertyName<String> LOCK_OWNER =
        new PropertyName<String>(PROPERTY_NAMESPACE, "lock-owner");

    /**
     * Returns the value of the {@link #LOCK_OWNER LOCK_OWNER} property as
     * defined by this proxy.
     * 
     * @return A String containing the login name of the user that has a lock on
     *         this record ("*" if the current user is not permitted to see
     *         other's login names). If no user has a lock on this record, a
     *         zero-length String is returned.
     * 
     * @throws WvcmException if this proxy does not define a value for the
     *             {@link #LOCK_OWNER LOCK_OWNER} property.
     */
    String getLockOwner() throws WvcmException;

    /**
     * Defines a new value for the {@link #LOCK_OWNER LOCK_OWNER} property of
     * this proxy.
     * <p>
     * Unlike the other writable properties of a record, updating the LOCK_OWNER
     * property <i>does not</i> require that an action be started or be in
     * progress. And if an action is started or is in progress, the record is
     * locked immediately and will remain locked only until the action is
     * delivered or reverted.
     * 
     * @param owner Either a zero-length String or a String containing the login
     *            name of the current user. Setting the value to a zero-length
     *            String will unlock the record. Setting a non-zero length value
     *            will lock the record.
     * 
     * @return This proxy.
     */
    CqRecord setLockOwner(String owner);
}
